"""
Linear regression example with spatial correlation model
----------------------------------------------------------------------------------------
The n data points (y1, y2, ..., yn) generated for this example are sampled from an
n-variate normal distribution with mean values given by yi = a * xi + b with a, b being
the model parameters and x1, x2, ..., xi, ..., xn being predefined spatial x-coordinates
ranging from 0 to 1. The data points (y1, y2, ..., yn) are not independent but
correlated. The corresponding covariance matrix is defined based on an exponential
correlation function parameterized by the const standard deviation sigma of the
n-variate normal distribution and a correlation length l_corr. Hence, the full model has
four parameters a, b, sigma, l_corr, all of which are inferred in this example using
via maximum likelihood estimation and via sampling using emcee and dynesty.
"""

# standard library
import unittest

# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines
from probeye.inference.scipy_.correlation_models import (
    SpatialExponentialCorrelationModel,
)


class TestProblem(unittest.TestCase):
    def test_spatial_correlation(
        self,
        n_steps: int = 200,
        n_initial_steps: int = 100,
        n_walkers: int = 20,
        plot: bool = False,
        show_progress: bool = False,
        run_scipy: bool = True,
        run_emcee: bool = True,
        run_torch: bool = False,
        run_dynesty: bool = True,
    ):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps
            Number of steps (samples) to run. Note that the default number is rather low
            just so the test does not take too long.
        n_initial_steps
            Number of steps for initial (burn-in) sampling.
        n_walkers
            Number of walkers used by the estimator.
        plot
            If True, the data and the posterior distributions are plotted. This is
            deactivated by default, so that the test does not stop until the generated
            plots are closed.
        show_progress
            If True, progress-bars will be shown, if available.
        run_scipy
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee
            If True, the problem is solved with the emcee solver. Otherwise, the emcee
            solver will not be used.
        run_torch
            If True, the problem is solved with the pyro/torch_ solver. Otherwise, the
            pyro/torch_ solver will not be used.
        run_dynesty
            If True, the problem is solved with the dynesty solver. Otherwise, the
            dynesty solver will not be used.
        """

        if run_torch:
            raise RuntimeError(
                "The pyro-solver is not available yet for forward models including "
                "correlations."
            )

        # ============================================================================ #
        #                              Set numeric values                              #
        # ============================================================================ #

        # 'true' value of a, and its normal prior parameters
        a_true = 2.5
        loc_a = 2.0
        scale_a = 1.0

        # 'true' value of b, and its normal prior parameters
        b_true = 1.7
        loc_b = 1.0
        scale_b = 1.0

        # 'true' value of noise sd, and its uniform prior parameters
        sigma = 0.5
        low_sigma = 0.1
        high_sigma = 0.8

        # 'true' value of correlation length, and its uniform prior parameters
        l_corr = 0.05
        low_l_corr = 0.001
        high_l_corr = 0.2

        # settings for the data generation
        n_experiments = 8
        n_points = 50
        seed = 1

        # ============================================================================ #
        #                           Define the Forward Model                           #
        # ============================================================================ #

        class LinearModel(ForwardModelBase):
            def response(self, inp: dict) -> dict:
                a = inp["a"]
                b = inp["b"]
                response = {}
                for os in self.output_sensors:
                    response[os.name] = a * os.x.flatten() + b
                return response

        # ============================================================================ #
        #                         Define the Inference Problem                         #
        # ============================================================================ #

        # initialize the inference problem with a useful name; note that the name will
        # only be stored as an attribute of the InferenceProblem and is not important
        # for the problem itself; can be useful when dealing with multiple problems
        problem = InferenceProblem("Linear regression with normal noise")

        # add all parameters to the problem; the first argument states the parameter's
        # global name (here: 'a', 'b' and 'sigma'); the second argument defines the
        # parameter type (three options: 'model' for parameter's of the forward model,
        # 'prior' for prior parameters and 'noise' for parameters of the noise model);
        # the 'info'-argument is a short description string used for logging, and the
        # tex-argument gives a tex-string of the parameter used for plotting; finally,
        # the prior-argument specifies the parameter's prior; note that this definition
        # of a prior will result in the initialization of constant parameters of type
        # 'prior' in the background
        problem.add_parameter(
            "a",
            "model",
            tex="$a$",
            info="Slope of the graph",
            prior=("normal", {"loc": loc_a, "scale": scale_a}),
        )
        problem.add_parameter(
            "b",
            "model",
            info="Intersection of graph with y-axis",
            tex="$b$",
            prior=("normal", {"loc": loc_b, "scale": scale_b}),
        )
        problem.add_parameter(
            "sigma",
            "noise",
            tex=r"$\sigma$",
            info="Std. dev, of 0-mean noise model",
            prior=("uniform", {"low": low_sigma, "high": high_sigma}),
        )
        problem.add_parameter(
            "l_corr",
            "noise",
            tex=r"$l_\mathrm{corr}$",
            info="Correlation length of correlation model",
            prior=("uniform", {"low": low_l_corr, "high": high_l_corr}),
        )

        # add the forward model to the problem; note that the first positional argument
        # [{'a': 'm'}, 'b'] passed to LinearModel defines the forward model's parameters
        # by name via a list with elements structured like {<global parameter name>:
        # <local parameter name>}; a global name is a name introduced by problem.
        # add_parameter, while a local name is a name used in the response-method of the
        # forward model class (see the class LinearModel above); note that the use of
        # the local parameter name 'm' for the global parameter 'a' is added here only
        # to highlight the possibility of this feature; it is not necessary at all here;
        # whenever forward model's parameter has a similar local and global name (which
        # should be the case most of the times), one doesn't have to use the verbose
        # notation  {<global parameter name>: <local parameter name>} but can instead
        # just write the parameter's (global=local) name, like it is done with the
        # forward model's parameter 'b' below
        x_test = np.linspace(0.0, 1.0, n_points)
        osensor = Sensor("y", x=x_test)
        linear_model = LinearModel(["a", "b"], [], [osensor])
        problem.add_forward_model("LinearModel", linear_model)

        # add the noise model to the problem
        noise_model = NormalNoiseModel(
            sensors=osensor,
            corr="x",
            corr_model="exp",
            prms_def=[{"sigma": "std"}, "l_corr"],
        )
        problem.add_noise_model(noise_model)

        # ============================================================================ #
        #                    Add test data to the Inference Problem                    #
        # ============================================================================ #

        # data-generation; first create the true values without noise; these true values
        # will be the mean values for sampling from a multivariate normal distribution
        np.random.seed(seed)
        y_true = linear_model({"a": a_true, "b": b_true})[osensor.name]

        # create the covariance matrix
        correlation_model = SpatialExponentialCorrelationModel(x=osensor.x)
        cov = correlation_model({"std": sigma, "l_corr": l_corr})

        # now generate the noisy test data including correlations; we assume here that
        # there are n_experiments test series
        for i in range(n_experiments):
            y_test = np.random.multivariate_normal(mean=y_true, cov=cov)
            problem.add_experiment(
                f"Test_{i}",
                fwd_model_name="LinearModel",
                sensor_values={osensor.name: y_test},
            )
            if plot:
                plt.scatter(
                    x_test, y_test, label=f"measured data (test {i+1})", s=10, zorder=10
                )
        # finish the plot
        if plot:
            plt.plot(x_test, y_true, label="true model", c="black", linewidth=3)
            plt.xlabel("x")
            plt.ylabel(osensor.name)
            plt.legend()
            plt.tight_layout()
            plt.draw()  # plt.draw() does not stop execution

        # ============================================================================ #
        #                    Solve problem with inference engine(s)                    #
        # ============================================================================ #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way
        true_values = {"a": a_true, "b": b_true, "sigma": sigma, "l_corr": l_corr}
        run_inference_engines(
            problem,
            true_values=true_values,
            n_steps=n_steps,
            n_initial_steps=n_initial_steps,
            n_walkers=n_walkers,
            plot=plot,
            show_progress=show_progress,
            run_scipy=run_scipy,
            run_emcee=run_emcee,
            run_torch=run_torch,
            run_dynesty=run_dynesty,
        )


if __name__ == "__main__":
    unittest.main()



# standard library imports
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.noise_model import NoiseModelBase
from probeye.definition.noise_model import NormalNoiseModel
from probeye.definition.sensor import Sensor


class TestProblem(unittest.TestCase):
    def test_init(self):
        # initialize an instance and check if everything is there
        noise_template = NoiseModelBase("normal", ["bias", "sigma"], Sensor("y"))
        self.assertEqual(noise_template.prms_def, {"bias": "bias", "sigma": "sigma"})
        self.assertEqual(noise_template.prms_dim, 2)
        noise_template = NoiseModelBase("normal", ["b", {"sd": "sigma"}], Sensor("y"))
        self.assertEqual(noise_template.prms_def, {"b": "b", "sd": "sigma"})
        self.assertEqual(noise_template.prms_dim, 2)
        self.assertEqual(noise_template.sensor_names, ["y"])
        self.assertEqual(noise_template.experiment_names, [])

    def test_add_experiment_names(self):
        # prepare the setup for the tests
        noise_template = NoiseModelBase("normal", ["bias", "sigma"], Sensor("y"))
        problem_experiments = {
            "Exp1": {"sensor_values": {"x": 1, "y": 2}, "forward_model": "TestModel"},
            "Exp2": {"sensor_values": {"x": 3, "y": 4}, "forward_model": "TestModel"},
            "Exp3": {"sensor_values": {"x": 5, "y": 6}, "forward_model": "TestModel"},
            "Exp4": {"sensor_values": {"x": 7, "z": 8}, "forward_model": "TestModel"},
            "Exp5": {"sensor_values": {"x": 7, "y": 8}, "forward_model": "TestModel"},
            "Exp6": {"sensor_values": {"x": 7, "y": 8}, "forward_model": "OtherModel"},
        }
        # the following is usually done automatically when adding the noise model to the
        # inference problem
        noise_template.problem_experiments = problem_experiments
        # perform some simple tests for valid usage
        noise_template.add_experiments("Exp1")
        self.assertEqual(noise_template.experiment_names, ["Exp1"])
        noise_template.add_experiments(["Exp2", "Exp3"])
        self.assertEqual(noise_template.experiment_names, ["Exp1", "Exp2", "Exp3"])
        # now check some invalid input arguments
        with self.assertRaises(RuntimeError):
            # add an experiment, that does not have the noise model's sensors
            noise_template.add_experiments("Exp4")
        with self.assertRaises(RuntimeError):
            # add experiments that refer to more than one forward model
            noise_template.add_experiments(["Exp5", "Exp6"])
        with self.assertRaises(RuntimeError):
            # adding the same experiment again
            noise_template.add_experiments("Exp1")

    def test_error(self):
        # prepare the setup for the tests
        noise_template = NoiseModelBase(
            "normal", ["bias", "sigma"], [Sensor("y1"), Sensor("y2")]
        )
        # in the NoiseModelBase class the error_function is just a dummy, so we need to
        # define it to be able to test the error-method
        noise_template.error_function = lambda d1, d2: {
            name: d1[name] - d2[name] for name in noise_template.sensor_names
        }
        model_response_dict = {
            "Exp1": {"x": 1, "y1": 3, "y2": -3},
            "Exp2": {"x": 3, "y1": 6, "y2": -6},
            "Exp3": {"x": 5, "y1": 9, "y2": -9},
        }
        problem_experiments = {
            "Exp1": {
                "sensor_values": {"x": 1, "y1": 2, "y2": -2},
                "forward_model": "TestModel",
            },
            "Exp2": {
                "sensor_values": {"x": 3, "y1": 4, "y2": -4},
                "forward_model": "TestModel",
            },
            "Exp3": {
                "sensor_values": {"x": 5, "y1": 6, "y2": -6},
                "forward_model": "TestModel",
            },
        }
        # the following is usually done automatically when adding the noise model to the
        # inference problem
        noise_template.problem_experiments = problem_experiments
        noise_template.add_experiments(["Exp1", "Exp2", "Exp3"])
        # now we can call the error-method
        computed_value = noise_template.error(model_response_dict)
        expected_value = {
            "y1": np.array([1.0, 2.0, 3.0]),
            "y2": np.array([-1.0, -2.0, -3.0]),
        }
        # due to the numpy arrays we have to loop over everything to compare
        for sensor_name, array in computed_value.items():
            for i in range(len(array)):
                self.assertEqual(array[i], expected_value[sensor_name][i])

    def test_normal_noise_init(self):
        # check valid initialization
        noise_model = NormalNoiseModel({"sigma": "std"}, Sensor("y"))
        self.assertTrue(noise_model.zero_mean)
        self.assertEqual(noise_model.noise_type, "additive")
        noise_model = NormalNoiseModel(
            {"sigma": "std"}, Sensor("y"), noise_type="multiplicative"
        )
        self.assertEqual(noise_model.noise_type, "multiplicative")
        noise_model = NormalNoiseModel(
            {"sigma": "std"}, Sensor("y"), noise_type="other"
        )
        self.assertEqual(noise_model.noise_type, "other")
        noise_model = NormalNoiseModel({"mean": "mean", "sigma": "std"}, Sensor("y"))
        self.assertTrue(not noise_model.zero_mean)
        # check invalid initialization
        with self.assertRaises(ValueError):
            # invalid noise_type given
            NormalNoiseModel({"sigma": "std"}, Sensor("y"), noise_type="invalid_type")
        with self.assertRaises(RuntimeError):
            # no 'std'-parameter given
            NormalNoiseModel({"sigma": "this should be 'std'"}, Sensor("y"))

    def test_normal_noise_error_function(self):
        # simply check that the error_function is set correctly when the noise model is
        # initialized
        ym_dict, ye_dict = {"y": 4}, {"y": 2}
        noise_model = NormalNoiseModel(
            {"sigma": "std"}, Sensor("y"), noise_type="additive"
        )
        computed_value = noise_model.error_function(ym_dict, ye_dict)
        expected_value = {"y": 2}
        self.assertEqual(computed_value, expected_value)
        noise_model = NormalNoiseModel(
            {"sigma": "std"}, Sensor("y"), noise_type="multiplicative"
        )
        computed_value = noise_model.error_function(ym_dict, ye_dict)
        expected_value = {"y": 1.0}
        self.assertEqual(computed_value, expected_value)
        # the 'other' error function is not pre-defined
        noise_model = NormalNoiseModel(
            {"sigma": "std"}, Sensor("y"), noise_type="other"
        )
        with self.assertRaises(NotImplementedError):
            noise_model.error_function(ym_dict, ye_dict)

    def test_normal_noise_error_function_additive(self):
        # check method for scalar values
        ym_dict, ye_dict = {"y": 4}, {"y": 2}
        noise_model = NormalNoiseModel({"sigma": "std"}, Sensor("y"))
        computed_value = noise_model.error_function_additive(ym_dict, ye_dict)
        expected_value = {"y": 2.0}
        self.assertEqual(computed_value, expected_value)
        # check method for numpy-arrays
        ym_dict, ye_dict = {"y": np.array([4, 5])}, {"y": np.array([2, 4])}
        noise_model = NormalNoiseModel({"sigma": "std"}, Sensor("y"))
        computed_value = noise_model.error_function_additive(ym_dict, ye_dict)
        expected_value = {"y": np.array([2, 1])}
        self.assertTrue(np.allclose(computed_value["y"], expected_value["y"]))

    def test_normal_noise_error_function_multiplicative(self):
        # check method for scalar values
        ym_dict, ye_dict = {"y": 4}, {"y": 2}
        noise_model = NormalNoiseModel({"sigma": "std"}, Sensor("y"))
        comp_value = noise_model.error_function_multiplicative(ym_dict, ye_dict)
        expected_value = {"y": 1.0}
        self.assertEqual(comp_value, expected_value)
        # check method for numpy-arrays
        ym_dict, ye_dict = {"y": np.array([4, 5])}, {"y": np.array([2, 4])}
        noise_model = NormalNoiseModel({"sigma": "std"}, Sensor("y"))
        comp_value = noise_model.error_function_multiplicative(ym_dict, ye_dict)
        expected_value = {"y": np.array([1.0, 0.25])}
        self.assertTrue(np.allclose(comp_value["y"], expected_value["y"]))

    def test_normal_noise_loglike_contribution(self):
        # the loglike-contribution is not defined in this general class
        noise_model = NormalNoiseModel({"sigma": "std"}, Sensor("y"))
        with self.assertRaises(NotImplementedError):
            noise_model.loglike_contribution({}, None)


if __name__ == "__main__":
    unittest.main()



"""
Inference problem with two forward models that share a common parameter
----------------------------------------------------------------------------------------
The first model equation is y = a * x + b with a, b being the model parameters and the
second model equation is y = alpha * x**2 + b where alpha is a new model parameter, and
b is the same model parameter as in the first model equation. Both forward models have
the same noise model with a normal zero-mean distribution where the standard deviation
is to be inferred. The problem is solved via max likelihood estimation and via sampling
using emcee, pyro and dynesty.
"""

# standard library imports
import unittest

# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines


class TestProblem(unittest.TestCase):
    def test_two_models(
        self,
        n_steps: int = 200,
        n_initial_steps: int = 100,
        n_walkers: int = 20,
        plot: bool = False,
        show_progress: bool = False,
        run_scipy: bool = True,
        run_emcee: bool = True,
        run_torch: bool = True,
        run_dynesty: bool = True,
    ):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps
            Number of steps (samples) to run. Note that the default number is rather low
            just so the test does not take too long.
        n_initial_steps
            Number of steps for initial (burn-in) sampling.
        n_walkers
            Number of walkers used by the estimator.
        plot
            If True, the data and the posterior distributions are plotted. This is
            deactivated by default, so that the test does not stop until the generated
            plots are closed.
        show_progress
            If True, progress-bars will be shown, if available.
        run_scipy
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee
            If True, the problem is solved with the emcee solver. Otherwise, the emcee
            solver will not be used.
        run_torch
            If True, the problem is solved with the pyro/torch_ solver. Otherwise, the
            pyro/torch_ solver will not be used.
        run_dynesty
            If True, the problem is solved with the dynesty solver. Otherwise, the
            dynesty solver will not be used.
        """

        # ============================================================================ #
        #                              Set numeric values                              #
        # ============================================================================ #

        # 'true' value of a, and its normal prior parameters
        a_true = 2.5
        loc_a = 2.0
        scale_a = 1.0

        # 'true' value of b, and its normal prior parameters
        b_true = 1.7
        loc_b = 1.0
        scale_b = 1.0

        # 'true' value of alpha, and its normal prior parameters
        alpha_true = 0.7
        loc_alpha = 2.0
        scale_alpha = 1.0

        # 'true' value of sigma, and its normal prior parameters
        sigma_true = 0.15
        low_sigma = 0.1
        high_sigma = 2.0

        # the number of generated experiment_names and seed for random numbers
        n_tests = 100
        seed = 1

        # ============================================================================ #
        #                          Define the Forward Models                           #
        # ============================================================================ #

        class LinearModel(ForwardModelBase):
            def response(self, inp: dict) -> dict:
                x = inp["x"]
                a = inp["a"]
                b = inp["b"]
                response = {}
                for os in self.output_sensors:
                    response[os.name] = a * x + b
                return response

        class QuadraticModel(ForwardModelBase):
            def response(self, inp: dict) -> dict:
                x = inp["x"]
                alpha = inp["alpha"]
                beta = inp["beta"]
                response = {}
                for os in self.output_sensors:
                    response[os.name] = alpha * x ** 2 + beta
                return response

        # ============================================================================ #
        #                         Define the Inference Problem                         #
        # ============================================================================ #

        # initialize the inference problem with a useful name
        problem = InferenceProblem("Two models with shared parameter and normal noise")

        # add all parameters to the problem
        problem.add_parameter(
            "a",
            "model",
            info="Slope of the graph in linear model",
            tex="$a$ (linear)",
            prior=("normal", {"loc": loc_a, "scale": scale_a}),
        )
        problem.add_parameter(
            "alpha",
            "model",
            info="Factor of quadratic term",
            tex=r"$\alpha$ (quad.)",
            prior=("normal", {"loc": loc_alpha, "scale": scale_alpha}),
        )
        problem.add_parameter(
            "b",
            "model",
            info="Intersection of graph with y-axis",
            tex="$b$ (shared)",
            prior=("normal", {"loc": loc_b, "scale": scale_b}),
        )
        problem.add_parameter(
            "sigma",
            "noise",
            tex=r"$\sigma$ (noise)",
            info="Std. deviation of zero-mean noise model",
            prior=("uniform", {"low": low_sigma, "high": high_sigma}),
        )

        # add the forward model to the problem
        isensor = Sensor("x")
        osensor_linear = Sensor("y_linear")
        osensor_quadratic = Sensor("y_quadratic")
        linear_model = LinearModel(["a", "b"], [isensor], [osensor_linear])
        problem.add_forward_model("LinearModel", linear_model)
        quadratic_model = QuadraticModel(
            ["alpha", {"b": "beta"}], [isensor], [osensor_quadratic]
        )
        problem.add_forward_model("QuadraticModel", quadratic_model)

        # add the noise model to the problem
        problem.add_noise_model(
            NormalNoiseModel(prms_def={"sigma": "std"}, sensors=osensor_linear)
        )
        problem.add_noise_model(
            NormalNoiseModel(prms_def={"sigma": "std"}, sensors=osensor_quadratic)
        )

        # ============================================================================ #
        #                    Add test data to the Inference Problem                    #
        # ============================================================================ #

        # data-generation; normal noise with constant variance around each point
        np.random.seed(seed)
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_linear_true = linear_model({isensor.name: x_test, "a": a_true, "b": b_true})[
            osensor_linear.name
        ]
        y_test_linear = np.random.normal(loc=y_linear_true, scale=sigma_true)
        y_quadratic_true = quadratic_model(
            {isensor.name: x_test, "alpha": alpha_true, "beta": b_true}
        )[osensor_quadratic.name]
        y_test_quadratic = np.random.normal(loc=y_quadratic_true, scale=sigma_true)

        # add the experimental data
        problem.add_experiment(
            f"TestSeries_linear",
            sensor_values={isensor.name: x_test, osensor_linear.name: y_test_linear},
            fwd_model_name="LinearModel",
        )
        problem.add_experiment(
            f"TestSeries_quadratic",
            sensor_values={
                isensor.name: x_test,
                osensor_quadratic.name: y_test_quadratic,
            },
            fwd_model_name="QuadraticModel",
        )

        # give problem overview
        problem.info()

        # plot the true and noisy data
        if plot:
            plt.scatter(
                x_test,
                y_test_linear,
                label="measured data (linear)",
                s=10,
                c="red",
                zorder=10,
            )
            plt.plot(x_test, y_linear_true, label="true (linear)", c="black")
            plt.scatter(
                x_test,
                y_test_quadratic,
                s=10,
                c="orange",
                zorder=10,
                label="measured data (quadratic)",
            )
            plt.plot(x_test, y_quadratic_true, label="true (quadratic)", c="blue")
            plt.xlabel(isensor.name)
            plt.ylabel(f"{osensor_linear.name}, {osensor_quadratic.name}")
            plt.legend()
            plt.tight_layout()
            plt.draw()  # does not stop execution

        # ============================================================================ #
        #                    Solve problem with inference engine(s)                    #
        # ============================================================================ #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way
        true_values = {
            "a": a_true,
            "alpha": alpha_true,
            "b": b_true,
            "sigma": sigma_true,
        }
        run_inference_engines(
            problem,
            true_values=true_values,
            n_steps=n_steps,
            n_initial_steps=n_initial_steps,
            n_walkers=n_walkers,
            plot=plot,
            show_progress=show_progress,
            run_scipy=run_scipy,
            run_emcee=run_emcee,
            run_torch=run_torch,
            run_dynesty=run_dynesty,
        )


if __name__ == "__main__":
    unittest.main()



# standard library
from typing import Union, List, Tuple, Optional

# third party imports
import numpy as np

# local imports
from probeye.subroutines import process_spatial_coordinates


class Sensor:
    """
    Base class for a sensor of the forward model. In its simplest form just a class with
    a 'name' attribute. Additional attributes for the measured quality (measurand) and
    the corresponding unit can be defined. If the sensors position(s) are important,
    they can be defined as attributes. Further attributes can be defined by the user by
    creating new classes derived from this one.
    """

    def __init__(
        self,
        name: str,
        measurand: str = "not defined",
        unit: str = "not defined",
        x: Union[float, int, np.ndarray, None] = None,
        y: Union[float, int, np.ndarray, None] = None,
        z: Union[float, int, np.ndarray, None] = None,
        coords: Optional[np.ndarray] = None,
        order: Tuple[str, ...] = ("x", "y", "z"),
    ):
        """
        Parameters
        ----------
        name
            The name of the sensor, e.g. 'Deflection-Sensor bottom-left'.
        measurand
            Defines what the sensor measures, e.g. 'deflection'.
        unit
            Defines what unit is associated with the sensor's measurements, e.g. 'mm'.
        x
            x-coordinate of the sensor. When given, the coords- argument must be None.
        y
            y-coordinate of the sensor. When given, the coords- argument must be None.
        z
            z-coordinate of the sensor. When given, the coords-argument must be None.
        coords
            Some or all of the coordinates x, y, z concatenated as an array. Each row
            corresponds to a constant coordinate. Which row corresponds to which
            coordinate is defined via the order-argument. When given, the arguments x, y
            and z must be None.
        order
            Only relevant when coords is given. Defines which row in coords corresponds
            to which coordinate. For example, order=('x', 'y', 'z') means that the 1st
            row are x-coordinates, the 2nd row are y-coords and the 3rd row are the
            z-coordinates.
        """

        # basic attributes
        self.name = name
        self.measurand = measurand
        self.unit = unit

        # translate the input to a coords-array
        self.coords, self._order = process_spatial_coordinates(
            x=x, y=y, z=z, coords=coords, order=order
        )

        # this contains the information which row contains which coordinate
        self.index_dict = {coord: i for i, coord in enumerate(self._order)}

    @property
    def x(self) -> Union[np.ndarray, None]:
        """Provides x-coords as attribute without copying them from coords."""
        return self.coords[self.index_dict["x"]] if "x" in self._order else None

    @property
    def y(self) -> Union[np.ndarray, None]:
        """Provides y-coords as attribute without copying them from coords."""
        return self.coords[self.index_dict["y"]] if "y" in self._order else None

    @property
    def z(self) -> Union[np.ndarray, None]:
        """Provides z-coords as attribute without copying them from coords."""
        return self.coords[self.index_dict["z"]] if "z" in self._order else None

    @property
    def order(self) -> List[str]:
        """Provides read-access to privat attribute self._order."""
        return self._order



# standard library
from typing import Union, Optional, Tuple

# third party imports
import numpy as np

# local imports
from probeye.subroutines import process_spatial_coordinates


class SpatialExponentialCorrelationModel:
    """
    Represents a spatial correlation model with an exponential kernel. It contains the
    functionality to compute the covariance matrix over a static (i.e. constant for all
    experiments) grid of coordinates in 1D, 2D or 3D.
    """

    def __init__(
        self,
        x: Union[int, float, np.ndarray, None] = None,
        y: Union[int, float, np.ndarray, None] = None,
        z: Union[int, float, np.ndarray, None] = None,
        coords: Optional[np.ndarray] = None,
        order: Tuple = ("x", "y", "z"),
    ):
        """
        Parameter
        ---------
        x
            Positional x-coordinate. When given, coords must be None.
        y
            Positional y-coordinate. When given, coords must be None.
        z
            Positional z-coordinate. When given, coords must be None.
        coords
            Some or all of the coordinates x, y, z concatenated as an array. Each row
            corresponds to one coordinate. For example, row 1 might contain all
            x-coordinates. Which row corresponds to which coordinate is defined via the
            order-argument. When the coords-argument is given, all 3 arguments x, y and
            z must be None.
        order
            Only relevant when coords is given. Defines which row in coords corresponds
            to which coordinate. For example, order=('x', 'y', 'z') means that the 1st
            row are x-coordinates, the 2nd row are y-coords and the 3rd row are the
            z-coordinates.
        """
        # translate the spatial input to a coords-array
        self.coords, self._order = process_spatial_coordinates(
            x=x, y=y, z=z, coords=coords, order=order
        )
        self.n_coords, self.n = self.coords.shape

        # on position (i, j) in self.distance array will be denoted the distance between
        # point i with coords[i, :] and point j with coords[j, :]
        distance_array = np.zeros((self.n, self.n))
        for i in range(self.n_coords):
            v = self.coords[i, :]
            v_in_columns = np.tile(v.reshape((self.n, -1)), self.n)
            v_in_rows = v_in_columns.transpose()  # each row is v
            distance_array += np.square(v_in_columns - v_in_rows)
        self.distance_array = np.sqrt(distance_array)

    @staticmethod
    def check_prms(prms: dict) -> bool:
        """
        Checks if the numeric values provided via prms are valid for computing the
        covariance matrix of this model.

        Parameters
        ----------
        prms
            Contains the names of the correlation model's parameters as keys and the
            corresponding numeric values as values.

        Returns
        -------
            True, when all values are valid. False otherwise.
        """
        # if no correlation is defined in the noise model, l_corr will not be provided
        # in the input for the log-likelihood contribution
        if "l_corr" in prms:
            if prms["l_corr"] <= 0:
                return False
        # sigma, however, must always be passed
        if prms["std"] <= 0:
            return False
        return True

    def __call__(self, prms: dict) -> np.ndarray:
        """
        Returns the covariance matrix based on the correlation model.

        Parameters
        ----------
        prms
            Contains the names of the correlation model's parameters as keys and the
            corresponding numeric values as values.

        Returns
        -------
            The covariance matrix based on the given parameters. The shape of
            this array is (self.n, self.n).
        """
        corr = np.exp(-self.distance_array / prms["l_corr"])
        return prms["std"] ** 2 * corr
